from botocore.stub import Stubber
from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch

from recruitment.agency import Job
from recruitment.agency import Coordinator
from recruitment.agency.resources import Broker
from tests.recruitment.agency import client


class JobTest(TestCase):

    broker = Broker.s3
    region = 'some-region-1'
    boto_client = client(broker.name, region)

    expected_create_target_response = {
        'ResponseMetadata': {
            'RequestId': 'MV004T58ZZ31T6TP',
            'HostId': 'pdeFdK6Uqq4yTyG4j+TdtrrG66PV8kc69qUHcj6cW7sIjpzjKYDFoy2acnO3TCAuVlX6qTbR1blyt+eM4C5nNg==',
            'HTTPStatusCode': 200,
            'HTTPHeaders': {
                'x-amz-id-2': 'pdeFdK6Uqq4yTyG4j+TdtrrG66PV8kc69qUHcj6cW7sIjpzjKYDFoy2acnO3TCAuVlX6qTbR1blyt+eM4C5nNg==',
                'x-amz-request-id': 'MV004T58ZZ31T6TP',
                'date': 'Sun, 30   Oct 2022 15:55:02 GMT',
                'location': '/test-bucket',
                'server': 'AmazonS3',
                'content-length': '0'
            },
            'RetryAttempts': 0
        },
        'Location': '/test-bucket'
    }

    @patch('boto3.client')
    def test_can_create_target(self, mock_boto_client):
        mock_boto_client.return_value = self.boto_client
        mock_commlink = MagicMock()
        mock_commlink.create_target.__name__ = 'create_target'
        mock_commlink.create_target.return_value = self.expected_create_target_response
        with Stubber(self.boto_client) as stubber:
            stubber.add_response(
                self.broker.interface['create_target'], self.expected_create_target_response
            )
            job = Job(Coordinator(mock_commlink))
            result = job.create_target(Bucket='test-bucket')

        self.assertTrue(result.successful)
        self.assertDictEqual(result.value, self.expected_create_target_response)



from actionpack.actions import RetryPolicy
from actionpack.actions import Write
from actionpack import partialaction
from enum import auto
from enum import Enum
from typing import Dict
from typing import Optional


class NaturalEnum(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name


class Broker(NaturalEnum):
    """A repository for declaring services and their interfaces"""

    logs = auto()
    s3 = auto()
    sns = auto()
    sqs = auto()
    kinesis = auto()

    @property
    def interface(self) -> Dict[str, Optional[str]]:
        send = 'send'
        create_target = 'create_target'
        receive = 'receive'
        methods_for = {
            Broker.logs: {receive: 'get_log_events'},
            Broker.s3: {
                create_target: 'create_bucket',
                receive: 'get_object',
                send: 'upload_fileobj',
            },
            Broker.sns: {send: 'publish', create_target: 'create_topic'},
            Broker.sqs: {
                create_target: 'create_queue',
                receive: 'receive_message',
                send: 'send_message',
            },
            Broker.kinesis: {send: 'put_record', create_target: 'create_stream'},
        }
        return methods_for[self]  # KeyError should be contextualized as NotImplementedError


class From(NaturalEnum):
    env = auto()
    file = auto()


class CloudProvider(NaturalEnum):
    AWS = auto()


#- Custom Actions ---------------------------->>>
Append = partialaction('Append', Write, append=True)
RecordedRetryPolicy = partialaction('RecordedRetryPolicy', RetryPolicy, should_record=True)
    


import boto3

from actionpack import Action
from actionpack.action import Result
from actionpack.actions import Call
from actionpack.utils import Closure
from botocore.exceptions import NoRegionError
from dataclasses import asdict
from dataclasses import dataclass
from functools import reduce
from os import environ as envvars
from pathlib import Path
from typing import Optional
from typing import TypeVar
from typing import Union

from recruitment.agency.resources import Broker
from recruitment.agency.resources import CloudProvider
from recruitment.agency.resources import From
from recruitment.agency.resources import RecordedRetryPolicy


T = TypeVar('T')
Reaction = Action

local_storage_dir = Path.home() / '.recruitment/agency/'
deadletters = local_storage_dir / 'deadletters'  # failures


@dataclass
class Config:
    """An object for conveying configuration info"""

    service_name: Union[str, Broker]
    region_name: Optional[str] = None
    access_key_id: Optional[str] = None
    secret_access_key: Optional[str] = None
    endpoint_url: Optional[str] = None

    @staticmethod
    def fromenv():
        return Config(
            service_name=envvars.get('AWS_SERVICE_NAME'),
            region_name=envvars.get('AWS_REGION_NAME'),
            access_key_id=envvars.get('AWS_ACCESS_KEY_ID'),
            secret_access_key=envvars.get('AWS_SECRET_ACCESS_KEY'),
            endpoint_url=envvars.get('AWS_ENDPOINT_URL')
        )

    def supplement(self, fromhere: str):
        fromwhere = From(fromhere)
        unset = {}
        if fromwhere == From.env:
            for k, v in asdict(self).items():
                if v is None:
                    unset[k] = envvars.get(f'{CloudProvider.AWS.value}_{k.upper()}')
        if fromwhere == From.file:
            raise NotImplementedError('Coming soon.')

        return Config(**{**asdict(self), **unset})

    def asfile(self, profile: str = 'default'):
        return f'[{profile}]\n{str(self)}'

    def __post_init__(self):
        if self.service_name is None:
            raise Config.AttributeDeclaredIncorrectly('Missing service_name.')

        if isinstance(self.service_name, Broker):
            self.service_name = self.service_name.value
        elif isinstance(self.service_name, str):
            self.service_name = Broker(self.service_name).value
        else:
            service_name_type = type(self.service_name).__name__
            raise Config.AttributeDeclaredIncorrectly(
                f'Service name must be a <Broker> or <str>. Received <{service_name_type}>.'
            )

    def __str__(self):
        """Produces a string compatible with the AWS CLI tool
        (see documentation for details:
         https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html)
        """
        return reduce(
            lambda a, b: f'{a}' + f'{b[0]}={b[1] if b[1] else ""}\n',
            asdict(self).items(),
            '',
        ).strip()

    class AttributeDeclaredIncorrectly(Exception):
        pass


class Commlink:
    """An object that hosts the Broker.interface"""

    def __init__(self, config: Config):
        self.broker = Broker(config.service_name)  # maybe redundant
        for alias, method in self.broker.interface.items():
            try:
                client = boto3.client(
                    service_name=config.service_name,
                    region_name=config.region_name,
                    aws_access_key_id=config.access_key_id,
                    aws_secret_access_key=config.secret_access_key,
                    endpoint_url=config.endpoint_url
                )
            except (ValueError, NoRegionError) as e:
                raise Commlink.FailedToInstantiate(given=config) from e
            setattr(self, alias, getattr(client, method))

    class FailedToInstantiate(Exception):
        def __init__(self, given: Config):
            redaction = '*' * 10
            redacted_config = Config(
                service_name=given.service_name,
                region_name=given.region_name,
                access_key_id=redaction,
                secret_access_key=redaction,
                endpoint_url=given.endpoint_url
            )
            super().__init__(str(redacted_config))


class Contingency:

    def __new__(cls, *args, **kwargs) -> T:
        instance = super().__new__(cls)
        param_names, assigned_param_names = ['reaction', 'max_retries'], []
        for param_name in param_names:
            param_value = kwargs.get(param_name)
            if param_value:
                assigned_param_names.append(param_name)
                setattr(instance, param_name, param_value)

        if not any(assigned_param_names):
            return cls.__call__(cls, *args, **kwargs)
        else:
            return instance

    def __call__(
        self,
        action: Action,
    ) -> RecordedRetryPolicy:
        return RecordedRetryPolicy(
            action=action,
            reaction=self.reaction if hasattr(self, 'reaction') else None,
            max_retries=self.max_retries if hasattr(self, 'max_retries') else 2  # retries
        )


class Coordinator:

    def __init__(
        self,
        commlink: Commlink,
        contingency: Optional[Contingency] = None
    ):
        self.commlink = commlink
        self.contingency = contingency

    def do(self, action: Action) -> Result[T]:
        if self.contingency:
            retry_policy = self.contingency(action=action)
            return retry_policy.perform(), retry_policy.attempts
        else:
            return action.perform()


class Job:

    def __init__(self, coordinator: Coordinator):
        self.coordinator = coordinator

    def create_target(self, *args, **kwargs):
        create_target = Call(Closure(self.coordinator.commlink.create_target, *args, **kwargs))
        return self.coordinator.do(create_target)

    def __repr__(self) -> str:
        name = self.__class__.__name__
        broker = self.coordinator.commlink.broker.name
        has_contingency = ':contingency' if self.coordinator.contingency else ''

        return f'<{name}:{broker}{has_contingency}>'


class Publisher(Job):
    """A namespace for publishing messages"""

    def publish(self, *args, **kwargs):
        send_communique = Call(Closure(self.coordinator.commlink.send, *args, **kwargs))
        return self.coordinator.do(send_communique)


class Consumer(Job):
    """A namespace for consuming messages"""

    def consume(self, *args, **kwargs):
        receive_communique = Call(Closure(self.coordinator.commlink.receive, *args, **kwargs))
        return self.coordinator.do(receive_communique)


class Agent:
    """A namespace for consuming and/or publishing messages"""

    def __init__(self, consumer: Consumer, publisher: Publisher):
        if not isinstance(consumer, Consumer):
            raise TypeError(f'{self.__class__.__name__} consumer must be of type {Consumer.__name__} not {type(consumer).__name__}')
        if not isinstance(publisher, Publisher):
            raise TypeError(f'{self.__class__.__name__} publisher must be of type {Publisher.__name__} not {type(publisher).__name__}')

        setattr(self, consumer.consume.__name__, consumer.consume)
        setattr(self, publisher.publish.__name__, publisher.publish)

        consumer_repr = repr(consumer).strip('<>')
        publisher_repr = repr(publisher).strip('<>')
        self.__repr = f'<{self.__class__.__name__}|{consumer_repr}|{publisher_repr}>'

    def __repr__(self) -> str:
        return self.__repr



from io import StringIO
from unittest import TestCase
from unittest.mock import patch

from actionpack import Action
from actionpack.actions import Call
from actionpack.actions import RetryPolicy
from actionpack.utils import Closure
from botocore.exceptions import ClientError
from botocore.stub import Stubber
from typing import Callable
from typing import Optional

from recruitment.agency import Agent
from recruitment.agency import Broker
from recruitment.agency import Commlink
from recruitment.agency import Config
from recruitment.agency import Consumer
from recruitment.agency import Contingency
from recruitment.agency import Coordinator
from recruitment.agency import Publisher
from recruitment.agency import Reaction
from tests.recruitment.agency import client
from tests.recruitment.agency import fake_credentials
from tests.recruitment.agency import uncloseable
from tests.recruitment.agency import write_to_deadletter_file



class AgentTest(TestCase):

    region = 'some-region-1'
    sns = client(Broker.sns.value, region)
    logs = client(Broker.logs.value, region)

    expected_publish_response = {'MessageId': '00000000-0000-0000-0000-000000000000'}
    expected_consume_response = {
        'events': [
            {
                'timestamp': 123,
                'message': 'string',
                'ingestionTime': 123
            },
        ],
        'nextForwardToken': 'string',
        'nextBackwardToken': 'string'
    }

    def publisher_provider(self, commlink: Commlink, contingency: Optional[Contingency] = None) -> Publisher:
        return Publisher(Coordinator(commlink=commlink, contingency=contingency))

    def consumer_provider(self, commlink: Commlink, contingency: Optional[Contingency] = None) -> Consumer:
        return Consumer(Coordinator(commlink=commlink, contingency=contingency))

    def commlink_provider(self, broker: Broker):
        return Commlink(Config(broker, **fake_credentials))

    def client_selector(self, *args, **kwargs):
        service_name = kwargs.get('service_name')
        if service_name == 'sns':
            return self.sns
        if service_name == 'logs':
            return self.logs

    @patch('boto3.client')
    def test_can_proceed_without_contingency(self, mock_boto_client):
        mock_boto_client.side_effect = self.client_selector
        with Stubber(self.logs) as logs_stubber, \
             Stubber(self.sns) as sns_stubber:
            sns_stubber.add_client_error(Broker.sns.interface['send'], '500')
            logs_stubber.add_client_error(Broker.logs.interface['receive'], '500')  # retry 2
            smith = Agent(
                consumer=self.consumer_provider(commlink=self.commlink_provider(Broker.logs)),
                publisher=self.publisher_provider(commlink=self.commlink_provider(Broker.sns))
            )
            consume_result = smith.consume(logGroupName='the-construct', logStreamName='the-training-program')
            publish_result = smith.publish(Message='Mr. Anderson...')

        self.assertFalse(consume_result.successful)
        self.assertIsInstance(consume_result.value, ClientError)

        self.assertFalse(publish_result.successful)
        self.assertIsInstance(publish_result.value, ClientError)

    @patch('boto3.client')
    def test_cannot_instantiate_agent_given_invalid_suboordinates(self, mock_boto_client):
        valid_consumer = self.consumer_provider(commlink=self.commlink_provider(Broker.logs))
        valid_publisher = self.publisher_provider(commlink=self.commlink_provider(Broker.sns))
        with self.assertRaises(TypeError):
            Agent(consumer='Invalid Suboordinate.', publisher=valid_publisher)
        with self.assertRaises(TypeError):
            Agent(consumer=valid_consumer, publisher='Invalid Suboordinate.')

    @patch('boto3.client')
    @patch('actionpack.actions.Write.perform')
    def test_can_retry_message_send(self, mock_write, mock_boto_client):
        mock_boto_client.side_effect = self.client_selector
        with Stubber(self.logs) as logs_stubber, \
             Stubber(self.sns) as sns_stubber:
            sns_stubber.add_client_error(Broker.sns.interface['send'], '500')  # attempt
            sns_stubber.add_client_error(Broker.sns.interface['send'], '500')  # retry 1
            sns_stubber.add_client_error(Broker.sns.interface['send'], '500')  # retry 2
            logs_stubber.add_client_error(Broker.logs.interface['receive'], '500')  # attempt
            logs_stubber.add_client_error(Broker.logs.interface['receive'], '500')  # retry 1
            logs_stubber.add_client_error(Broker.logs.interface['receive'], '500')  # retry 2
            smith = Agent(
                consumer=self.consumer_provider(
                    commlink=self.commlink_provider(Broker.logs),
                    contingency=Contingency
                ),
                publisher=self.publisher_provider(
                    commlink=self.commlink_provider(Broker.sns),
                    contingency=Contingency
                )
            )
            publish_result, publish_attempts = smith.publish(Message='Mr. Anderson...')
            consume_result, consume_attempts = smith.consume(logGroupName='the construct', logStreamName='the-training-program')

        self.assertFalse(publish_result.successful)
        self.assertIsInstance(publish_result.value, RetryPolicy.Expired)
        self.assertEqual(len(publish_attempts), 3)
        for attempt in publish_attempts:
            self.assertIsInstance(attempt.value, ClientError)

        self.assertFalse(consume_result.successful)
        self.assertIsInstance(consume_result.value, RetryPolicy.Expired)
        self.assertEqual(len(consume_attempts), 3)
        for attempt in consume_attempts:
            self.assertIsInstance(attempt.value, ClientError)

    @patch('boto3.client')
    @patch('actionpack.actions.Write.perform')
    def test_publish_can_eventually_succeed(self, mock_write, mock_boto_client):
        mock_boto_client.side_effect = self.client_selector
        consumer_max_retries = 1
        with Stubber(self.logs) as logs_stubber, \
             Stubber(self.sns) as sns_stubber:
            sns_stubber.add_client_error(Broker.sns.interface['send'], '500')  # attempt
            sns_stubber.add_client_error(Broker.sns.interface['send'], '500')  # retry 1
            sns_stubber.add_response(Broker.sns.interface['send'], self.expected_publish_response)  # retry 2
            logs_stubber.add_client_error(Broker.logs.interface['receive'], '500')  # attempt
            logs_stubber.add_response(Broker.logs.interface['receive'], self.expected_consume_response)  # retry 1
            smith = Agent(
                consumer=self.consumer_provider(
                    commlink=self.commlink_provider(Broker.logs),
                    contingency=Contingency(max_retries=consumer_max_retries)
                ),
                publisher=self.publisher_provider(
                    commlink=self.commlink_provider(Broker.sns),
                    contingency=Contingency
                )
            )
            publish_result, publish_attempts = smith.publish(Message='Mr. Anderson...')
            consume_result, consume_attempts = smith.consume(logGroupName='the-construct', logStreamName='the-training-program')

        self.assertTrue(publish_result.successful)
        self.assertEqual(publish_result.value, self.expected_publish_response)
        self.assertEqual(len(publish_attempts), 3)
        self.assertIsInstance(publish_attempts.pop().value, type(self.expected_publish_response))
        for attempt in publish_attempts:
            self.assertIsInstance(attempt.value, ClientError)

        self.assertTrue(consume_result.successful)
        self.assertEqual(consume_result.value, self.expected_consume_response)
        self.assertEqual(len(consume_attempts), consumer_max_retries + 1)
        self.assertIsInstance(consume_attempts.pop().value, type(self.expected_consume_response))
        for attempt in consume_attempts:
            self.assertIsInstance(attempt.value, ClientError)

    @patch('boto3.client')
    @patch('actionpack.actions.Write.perform')
    def test_can_define_failure_callback(self, mock_write, mock_boto_client):
        vessel = []
        contents = 'water'

        def fill(vessel: list, contents):
            vessel.append(contents)
            return vessel

        callback = Call(Closure(fill, vessel, contents))

        self.assertFalse(vessel)  # confirms the vessel has yet to be filled
        mock_boto_client.side_effect = self.client_selector
        publisher_max_retries, consumer_max_retries = 2, 1
        with Stubber(self.logs) as logs_stubber, \
             Stubber(self.sns) as sns_stubber:
            [sns_stubber.add_client_error(Broker.sns.interface['send'], '500') for _ in range(publisher_max_retries + 1)]
            [logs_stubber.add_client_error(Broker.logs.interface['receive'], '500') for _ in range(consumer_max_retries + 1)]
            smith = Agent(
                consumer=self.consumer_provider(
                    commlink=self.commlink_provider(Broker.logs),
                    contingency=Contingency(
                        reaction=callback,  # called if the RetryPolicy fails for any reason
                        max_retries=consumer_max_retries
                    )
                ),
                publisher=self.publisher_provider(
                    commlink=self.commlink_provider(Broker.sns),
                    contingency=Contingency
                )
            )
            publish_result, publish_attempts = smith.publish(Message='Mr. Anderson...')
            consume_result, consume_attempts = smith.consume(logGroupName='the-construct', logStreamName='the-training-program')

        self.assertFalse(publish_result.successful)
        self.assertIsInstance(publish_result.value, RetryPolicy.Expired)
        self.assertEqual(len(publish_attempts), publisher_max_retries + 1)
        for attempt in publish_attempts:
            self.assertIsInstance(attempt.value, ClientError)

        self.assertFalse(consume_result.successful)
        self.assertIsInstance(consume_result.value, RetryPolicy.Expired)
        self.assertEqual(len(consume_attempts), consumer_max_retries + 1)
        for attempt in consume_attempts:
            self.assertIsInstance(attempt.value, ClientError)

        self.assertEqual(vessel, [contents])  # evidence the fill callback was called

    @patch('boto3.client')
    @patch('pathlib.Path.open')
    def test_can_write_deadletter(self, mock_file, mock_boto_client):
        mock_boto_client.side_effect = self.client_selector
        with uncloseable(StringIO()) as buffer:
            mock_file.return_value = buffer

            with Stubber(self.sns) as sns_stubber:
                sns_stubber.add_client_error(Broker.sns.interface['send'], '500')
                sns_stubber.add_client_error(Broker.sns.interface['send'], '500')
                sns_stubber.add_client_error(Broker.sns.interface['send'], '500')
                smith = Agent(
                    consumer=self.consumer_provider(self.commlink_provider(Broker.logs)),
                    publisher=self.publisher_provider(
                        commlink=self.commlink_provider(Broker.sns),
                        contingency=Contingency(reaction=write_to_deadletter_file)
                    )
                )
                result, attempts = smith.publish(Message='Mr. Anderson...')

        buffer_contents: str = buffer.read()
        self.assertTrue(buffer_contents.startswith('->'))
        self.assertTrue(buffer_contents.endswith('failed\n'))

        self.assertFalse(result.successful)
        self.assertIsInstance(result.value, RetryPolicy.Expired)
        self.assertEqual(len(attempts), 3)
        for attempt in attempts:
            self.assertIsInstance(attempt.value, ClientError)
